import os

from django.conf import settings
from django.core import exceptions
from django.utils.importlib import import_module
from django_nose.plugin import DjangoSetUpPlugin, ResultPlugin, TestReorderer
from django_nose.runner import NoseTestSuiteRunner

import nose.core

from plugin import RedRoverFilter


def _get_plugins_from_settings():
    plugins = (
        list(getattr(settings, 'NOSE_PLUGINS', [])) + [
            'django_nose.plugin.TestReorderer',
            'redrover.plugin.RedRoverFilter'])

    for plug_path in plugins:
        try:
            dot = plug_path.rindex('.')
        except ValueError:
            raise exceptions.ImproperlyConfigured(
                "%s isn't a Nose plugin module" % plug_path)
        p_mod, p_classname = plug_path[:dot], plug_path[dot + 1:]

        try:
            mod = import_module(p_mod)
        except ImportError, e:
            raise exceptions.ImproperlyConfigured(
                'Error importing Nose plugin module %s: "%s"' % (p_mod, e))

        try:
            p_class = getattr(mod, p_classname)
        except AttributeError:
            raise exceptions.ImproperlyConfigured(
                'Nose plugin module "%s" does not define a "%s"' %
                (p_mod, p_classname))

        yield p_class()


class RedRoverRunner(NoseTestSuiteRunner):

    def __init__(self, *args, **kwargs):
        if 'NOSE_REDROVER' not in os.environ:
            os.environ['NOSE_REDROVER'] = '1'
        super(RedRoverRunner, self).__init__(*args, **kwargs)

    def run_suite(self, nose_argv):
        result_plugin = ResultPlugin()
        plugins_to_add = [
            DjangoSetUpPlugin(self),
            result_plugin,
            TestReorderer(),
            RedRoverFilter()]

        for plugin in _get_plugins_from_settings():
            plugins_to_add.append(plugin)

        nose.core.TestProgram(
            argv=nose_argv, exit=False, addplugins=plugins_to_add)
        return result_plugin.result



from be_exactly_assertion import BeExactlyAssertion
from be_valid_assertion import BeValidAssertion
from contain_assertion import ContainAssertion
from equal_assertion import EqualAssertion
from have_selector_assertion import HaveSelectorAssertion
from have_text_assertion import HaveTextAssertion
from respond_to_assertion import RespondToAssertion


__all__ = [
    'BeExactlyAssertion', 'BeValidAssertion', 'ContainAssertion',
    'EqualAssertion', 'HaveSelectorAssertion', 'HaveTextAssertion',
    'RespondToAssertion']



class ExpectationError(AssertionError):
    pass


def _expectation(_expected_outcome, _action_class, *_args, **_kwargs):

    class Expectation(object):

        def __init__(self, action_class, *args, **kwargs):
            self.action = action_class(*args, **kwargs)

        def __enter__(self):
            return self.action.enter()

        def __exit__(self, type, value, traceback):
            if self.action.exit() != _expected_outcome:
                raise ExpectationError(self.action.message)
                return False
            return True

    return Expectation(_action_class, *_args, **_kwargs)


class Expect(object):

    @classmethod
    def to(cls, action_class, *args, **kwargs):
        return _expectation(True, action_class, *args, **kwargs)

    @classmethod
    def not_to(cls, action_class, *args, **kwargs):
        return _expectation(False, action_class, *args, **kwargs)



from django.core.urlresolvers import NoReverseMatch, reverse
from django.db.models import Model as DjangoModel
from django.utils.encoding import iri_to_uri


def get_url(obj, *args, **kwargs):
    """Try to infer a URL from an ambiguous object."""
    if isinstance(obj, DjangoModel):
        return obj.get_absolute_url()
    obj = iri_to_uri(obj)
    if obj.startswith('http'):
        return obj
    try:
        return reverse(obj, *args, **kwargs)
    except NoReverseMatch:
        return obj
    except Exception, msg:
        raise RuntimeError(msg)



from base import BaseExpectation


class ChangeExpectation(BaseExpectation):
    """Determine if the result of a callable changes."""

    def __init__(self, subject, by=None):
        if callable(subject):
            self.subject = subject
        else:
            raise RuntimeError(
                "The subject of a change expectation must be a callable.")
        self.delta = by

    def enter(self):
        self.original_value = self.subject()
        return self.original_value

    def exit(self):
        self.new_value = self.subject()
        if self.delta:
            self.passes = abs(
                self.original_value - self.new_value) == self.delta
        else:
            self.passes = self.original_value != self.new_value
        return self.passes

    @property
    def message(self):
        if self.passes:
            if self.delta:
                msg = '{subject} changed by {delta}'
            else:
                msg = '{subject} changed'
        else:
            if self.delta:
                msg = ('{subject} did not change by {delta} '
                       '(it changed by {actual})')
            else:
                msg = '{subject} did not change'

        return msg.format(
            subject=repr(self.subject),
            delta=repr(self.delta),
            actual=abs(self.original_value - self.new_value))



import functools
from types import FunctionType

from django.test import LiveServerTestCase, TestCase
import splinter

from subject import get_splinter_actions, _subject, BROWSER


def _get_extra_context(instance, extra_context={}):
    subject_name = getattr(instance, 'subject', None)

    if subject_name:
        extra_context = {
            'it': _subject(instance)(subject_name),
            'its': _subject(instance, parent_name=subject_name)}

        if subject_name == BROWSER:
            extra_context.update(get_splinter_actions(instance))

    return extra_context


def before(func):
    """
    The only way to get `setUp` to behave the way we want is to
    decorate it. So might as well make it look intentional :)

    """
    @functools.wraps(func)
    def setUp(instance, *args, **kwargs):
        func.__globals__.update(_get_extra_context(instance))
        return func(instance, *args, **kwargs)

    return setUp


def describe(func):
    """
    This more or less is the "magic" of RedRover.  With the `describe`
    decorator, a test is equipped with the keywords 'it', 'its', and
    if applicable, 'visit'.

    """
    @functools.wraps(func)
    def test_func(instance, *args, **kwargs):
        func.__globals__.update(_get_extra_context(instance))
        return func(instance, *args, **kwargs)

    return test_func


def _is_protected(item):
    """Filter out things we shouldn't be messing with."""
    return ('__' in item or item in dir(LiveServerTestCase))


def _redrover_klass():
    """
    Prefix test methods with 'test_' so they are discoverable by the
    test runner.

    """
    class DiscoverableTests(type):

        def __new__(cls, name, bases, dct):
            for attr, value in dct.iteritems():
                if not _is_protected(attr) and isinstance(value, FunctionType):
                    value.__name__ = 'test_%s' % value.__name__
            return super(
                DiscoverableTests, cls).__new__(cls, name, bases, dct)

    return DiscoverableTests


class RedRoverTest(TestCase):

    __metaclass__ = _redrover_klass()


class RedRoverLiveTest(LiveServerTestCase):

    def __init__(self, *args, **kwargs):
        if getattr(self, 'subject', None) == BROWSER:
            setattr(self, BROWSER, splinter.Browser('zope.testbrowser'))
            self.setUp.__globals__.update(get_splinter_actions(self))
        super(RedRoverLiveTest, self).__init__(*args, **kwargs)

    __metaclass__ = _redrover_klass()


class RedRoverHelper(object):

    def __init__(self, parent):
        for i in dir(parent):
            if not i.startswith('__'):
                setattr(self, i, getattr(parent, i))
        if self.subject == BROWSER:
            self.entry.__globals__.update(get_splinter_actions(parent))
            self.exit.__globals__.update(get_splinter_actions(parent))

    def __enter__(self):
        return self.entry()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            raise exc_type(exc_value)
        return self.exit()

    def entry(self):
        pass

    def exit(self):
        pass



from django.core.exceptions import ValidationError
from django.db import models

from base import BaseAssertion


class BeValidAssertion(BaseAssertion):
    """Determine if a Django Model is valid."""

    def __init__(self, subject):
        if not isinstance(subject, models.Model):
            raise ValueError(
                '`be_valid` must be called on a Django Model instance.')

        self.subject = subject
        self.passes = True
        try:
            subject.full_clean()
        except ValidationError, e:
            self.passes = False
            self.error_messages = e.message_dict

    @property
    def message(self):
        if self.passes:
            msg = '{subject} is valid'
            return msg.format(subject=repr(self.subject))
        else:
            msg = '{subject} is not valid.  {message}'
            errors = []
            for i in self.error_messages:
                errors += [(i, ' '.join(self.error_messages[i]))]
            return msg.format(
                subject=repr(self.subject),
                message="\033[0m" + ''.join(["%s: %s" % i for i in errors]))



from base import BaseAssertion


class EqualAssertion(BaseAssertion):
    """Compare the equivalence of two objects."""

    def __init__(self, subject, other):
        self.subject = subject
        self.other = other
        self.passes = subject == other

    @property
    def message(self):
        if self.passes:
            msg = '{subject} equals {other}'
        else:
            msg = '{subject} does not equal {other}'

        return msg.format(
            subject=repr(self.subject),
            other=repr(self.other))



import os
import re

import termstyle
from rednose import RedNose


class RedRoverFilter(RedNose):
    """
    Suppress the last traceback entry since that will always be raised
    by redrover.

    """
    env_opt = 'NOSE_REDROVER'
    env_opt_color = 'NOSE_REDROVER_COLOR'
    score = 199

    def options(self, parser, env=os.environ):
        global REDNOSE_DEBUG
        rednose_on = bool(env.get(self.env_opt, False))
        rednose_color = env.get(self.env_opt_color, 'auto')
        REDNOSE_DEBUG = bool(env.get('REDNOSE_DEBUG', False))

        parser.add_option(
            "--redrover-rednose", action="store_true",
            default=rednose_on, dest="redrover_rednose",
            help="enable colour output (alternatively, set $%s=1)" % (
                self.env_opt,))
        parser.add_option(
            "--no-redrover-color", action="store_false",
            dest="redrover_rednose",
            help="disable colour output")
        parser.add_option(
            "--force-redrover-color", action="store_const",
            dest='redrover_rednose_color',
            default=rednose_color,
            const='force',
            help=("force colour output when not using a TTY " +
                  "(alternatively, set $%s=force)" % (self.env_opt_color,)))
        parser.add_option(
            "--redrover-immediate", action="store_true",
            default=False,
            help=("print errors and failures as they happen, "
                  "as well as at the end"))

    def configure(self, options, conf):
        if options.redrover_rednose:
            self.enabled = True
            termstyle_init = {
                'force': termstyle.enable,
                'off': termstyle.disable
            }.get(options.redrover_rednose_color, termstyle.auto)
            termstyle_init()

            self.immediate = options.immediate
            self.verbose = options.verbosity >= 2

    def _fmt_traceback(self, trace):
        """format a traceback"""
        ret = []
        ret.append(termstyle.default("   Traceback (most recent call last):"))
        current_trace = trace
        while current_trace is not None:
            line = self._file_line(current_trace)
            if line is not None and not re.search(r'redrover/\w+\.pyc?', line):
                ret.append(line)
            current_trace = current_trace.tb_next
        return '\n'.join(ret)



from urlparse import urlparse

from base import BaseAssertion


class HaveSelectorAssertion(BaseAssertion):
    """
    Determine if given selector exists in a splinter Browser document.

    """
    def __init__(self, browser, selector, **kwargs):
        if not hasattr(browser, 'html'):
            raise ValueError(
                '`have_selector` must be called on a page/Browser instance.')
        self.url = urlparse(browser.url).path
        self.selector = selector
        self.extras = kwargs
        elements = browser.find_by_css(selector)
        # Special requests
        for k in kwargs:
            elements = [
                el for el in elements if getattr(el, k).strip() == kwargs[k]]

        self.passes = bool(elements)

    @property
    def message(self):
        if self.passes:
            msg = '{selector} selector{extras} exists in {url}'
        else:
            msg = '{selector} selector{extras} does not exist in {url}'

        if self.extras:
            extras = ' with the %s' % ', and the '.join(
                ['%s "%s"' % (k, self.extras[k]) for k in self.extras])
        else:
            extras = ''

        return msg.format(
            selector=repr(self.selector),
            url=self.url,
            extras=extras)



from urlparse import urlparse

from base import BaseAssertion


class HaveTextAssertion(BaseAssertion):
    """
    Determine if given text appears in an splinter Browser document.

    """
    def __init__(self, browser, text):
        if not hasattr(browser, 'html'):
            raise ValueError(
                '`have_text` must be called on a page/Browser instance.')
        self.url = urlparse(browser.url).path
        self.text = text
        self.passes = text in browser.html

    @property
    def message(self):
        if self.passes:
            msg = '{text} appears in {url}'
        else:
            msg = '{text} does not appear in {url}'

        return msg.format(
            text=repr(self.text),
            url=self.url)



from httplib import InvalidURL
from urllib2 import HTTPError
from urlparse import urlparse

from splinter.exceptions import ElementDoesNotExist
from zope.testbrowser.browser import Browser as ZopeTestBrowser
from utils import get_url


class ElementNotFound(AssertionError):
    pass


BROWSER = '_browser'


def _splinter_action(_action, _parent):
    """
    All the various splinter actions we support.  (and a few others)

    """
    _browser = getattr(_parent, BROWSER)

    # Finders

    def find_button(query):
        # Grab all buttons
        buttons = _browser.find_by_tag('button')
        # First try to find the right one matching text
        matching_text = filter(lambda e: e.text.strip() == query, buttons)
        if matching_text:
            return matching_text[0]
        # Try by CSS selector
        buttons = _browser.find_by_css(query)
        if buttons:
            return buttons.first
        raise ElementNotFound('Could not find a button matching "%s"' % query)

    def find_link(query):
        # First we'll try by text, then by CSS selector
        links = _browser.find_link_by_text(query)
        if links:
            return links.first
        links = _browser.find_by_css('a').find_by_css(query)
        if links:
            return links.first
        raise ElementNotFound('Could not find a link matching "%s"' % query)

    def find_input(query):
        #
        # NO LONGER USED
        #
        # Try by label text first
        labels = _browser.find_by_tag('label')
        matching_text = filter(lambda e: e.text == query, labels)
        if matching_text:
            input_id = matching_text[0]._element.for_element.attrib['id']
            try:
                return _browser.find_by_id(input_id).first
            except ElementDoesNotExist:
                raise ElementNotFound(
                    'Found label "%s", but no matching input' % query)
        # Try by name
        try:
            inputs = _browser.find_by_name(query)
        except LookupError:
            # Zope Testbrowser throws this if the query is not a valid name.
            inputs = None
        if inputs:
            return inputs.first
        # Try by CSS selector
        inputs = _browser.find_by_css(query)
        if inputs:
            return inputs.first
        raise ElementNotFound('Could not find an input matching "%s"' % query)

    # Actions

    def fill_in(query, value):
        try:
            control = _browser._browser.getControl(query, index=0)
        except LookupError:
            try:
                control = _browser._browser.getControl(name=query)
            except LookupError:
                control = None

        if control:
            control.value = value
        else:
            input_field = find_input(query)
            input_field.fill(value)

    def click_on(query):
        try:
            # First try it the easy way (if the button is actually a
            # submit input).
            control = _browser._browser.getControl(query)
        except LookupError:
            button = find_button(query)
            try:
                control = _browser._browser.getControl(
                    name=button._element.attrib['name'])
            except KeyError:
                raise KeyError(
                    'Buttons must have a unique `name` attribute to be used.')
        return control.click()

    def select(query, option):
        try:
            control = _browser._browser.getControl(query)
        except LookupError:
            control = _browser._browser.getControl(name=query)
        control.value = [option]

    def visit(location, *args, **kwargs):
        try:
            url = get_url(location, *args, **kwargs)
            if url.startswith('/'):
                base_url = _parent.live_server_url
                return _browser.visit('%s%s' % (base_url, url))
            return _browser.visit(location)
        except HTTPError:
            raise
        except:
            raise InvalidURL(
                'Could not determine a reliable URL from "%s"' % location)

    def click_link(*args, **kwargs):
        control = _browser._browser.getLink(*args, **kwargs)
        return control.click()

    return locals().get(_action)


def get_splinter_actions(instance):
    """
    A dictionary of all supported Splinter actions plus a few of my own
    set up against the given test-class instance.

    """
    finders = ['find_button', 'find_link', 'find_input']
    actions = ['click_link', 'click_on', 'fill_in', 'select', 'visit']

    all_actions = {
        k: _splinter_action(k, instance) for k in finders + actions}
    all_actions['current_path'] = _subject(
        instance._browser, modifier=lambda url: urlparse(url).path)('url')
    return all_actions


def _subject(instance, parent_name=None, modifier=None):
    """
    The Subject object references a particular attribute and allows the
    developer to check it for various behaivors and conditions.

    """
    class Subject(object):

        def __init__(self, name):
            self.name = name
            self.instance = instance  # For debugging

        def should(self, cls, *args, **kwargs):
            assertion = cls(self.value, *args, **kwargs)
            if not assertion.passes:
                raise AssertionError(assertion.message)
            return True

        def should_not(self, cls, *args, **kwargs):
            assertion = cls(self.value, *args, **kwargs)
            if assertion.passes:
                raise AssertionError(assertion.message)
            return True

        @property
        def value(self):
            if parent_name:
                parent = getattr(instance, parent_name)
                if self.name == 'path':
                    if isinstance(
                            getattr(parent, BROWSER, None), ZopeTestBrowser):
                        _value = urlparse(parent.url).path
                else:
                    _value = getattr(parent, self.name)
            else:
                _value = getattr(instance, self.name)
            if modifier:
                _value = modifier(_value)
            return _value

        def __str__(self):
            return str(self.value)

        def __repr__(self):
            return self.__str__()

        def __lt__(self, other):
            return self.value < other

        def __le__(self, other):
            return self.value <= other

        def __eq__(self, other):
            return self.value == other

        def __ne__(self, other):
            return self.value != other

        def __gt__(self, other):
            return self.value > other

        def __ge__(self, other):
            return self.value >= other

    return Subject



from base import BaseAssertion


class ContainAssertion(BaseAssertion):
    """Determine if one object exists within another."""

    def __init__(self, subject, other):
        self.subject = subject
        self.other = other
        self.passes = other in subject

    @property
    def message(self):
        if self.passes:
            msg = '{other} is in {subject}'
        else:
            msg = '{other} is not in {subject}'

        return msg.format(
            subject=repr(self.subject),
            other=repr(self.other))



from django.db import models

from base import BaseAssertion


class RespondToAssertion(BaseAssertion):
    """
    Determine if a Django Model has a particular attribute or method.

    """
    def __init__(self, subject, attr):
        if not isinstance(subject, models.Model):
            raise ValueError(
                '`respond_to` must be called on a Django Model instance.')

        self.subject = subject
        self.attribute = attr
        fields = [f.__dict__['name'] for f in subject._meta.fields]
        attributes = dir(subject)
        self.passes = attr in fields + attributes

    @property
    def message(self):
        if self.passes:
            msg = '{subject} responds to {attribute}'
        else:
            msg = '{subject} does not respond to {attribute}'

        return msg.format(
            subject=repr(self.subject),
            attribute=repr(self.attribute))
